#!/bin/sh
# Shutdown dialog 
# Copyright (C) James Budiono 2012
# License: GNU GPL Version 3 or later
#
# 1.0 - initial release
#
# What it does: create savefile, set new hostname
# Called by wmexit

### configuration variables
APPTITLE="Create Save File"
ENCTYPE=aes
DOWNLOAD_FOLDER=/root/spot/Downloads 
. $BOOTSTATE_PATH					 # BOOTSTATE_PATH exported by /etc/profile
[ "$SAVEFILE_MOUNT" ] && exit		 # already have a savefile

### runtime variables
step=1						# current wizard step
savedevice=					# chosen save device (cannot be empty)
savedevicetype=				# filesystem type of the save device
savefile=fd64save.ext4		# chosen savefile name and path (cannot be empty)
savepath=					# path component of savefile, calculated from savefile
savesize=512				# chosen size of savefile, in MB (cannot be empty)
savefs=						# chosen filesystem of savefile (cannot be empty)
encrypted=					# chosen encryption type - none or $ENCTYPE
savepassword=				# chosen password for encryption (cannot be empty)
savefile_mount=				# temporary mount point for savefile
movedownload=on				# whether we move download folder
savedownload=				# the symlink to the new download folder
hostname=$(hostname)		# new hostname
rootpassword=woofwoof		# new root password

### step movements
go_step() { step=$1; }
go_next() { step=$((step + 1)); }
go_prev() { step=$((step - 1)); }
go_exit() { step=exit; }

prev_action=
prev_step=
next_wizard_step() {
	local action=$?
	[ "$1" ] && action=$1
	prev_step=$step
	case $action in
		0) go_next ;;				# Xdialog ok
		3) go_prev ;;				# Xdialog previous
		1|255) go_step warning ;;	# Xdialog Esc/window kill
	esac
	prev_action=$action
}


### actual steps here
introduction() {
	local msg=$(cat << EOF
You are running in RAM. At this moment, none of your data in this session is saved.	

If you want to save your data, you need to create a savefile. Do you want to create a savefile now?
EOF
)
	Xdialog --fill --title "$APPTITLE" --yesno "$msg" 0 0
	next_wizard_step
}

### warn we are about to exit without creating savefile
data_loss_warning() {
	local msg=$(cat <<EOF
You click "Cancel" during the savefile creation process. If you cancel this process, none of your data in this session will be saved.

Are you sure you want to stop this process, and lose all your data in this session?
EOF
)
	if Xdialog --fill --title "$APPTITLE" --yesno "$msg" 0 0; then go_exit
	else go_step $prev_step
	fi
}

### get the device / partition to save
choose_device() {
	local FS_TYPE label size msg="Choose the device/partition to save to."
	
	local items=$(
	while read p; do
		FS_TYPE=$(guess_fstype /dev/$p)
		case "$FS_TYPE" in
			unknown|"") ;;
			*) #label=$(disktype /dev/$p | sed -n '/Volume name/ { s/.*"\(.*\)".*/\1/; p}')
			   label=$(busybox blkid /dev/$p | sed -n '/LABEL/ {s/.*LABEL="\([^"]*\)".*/\1/; p}')
			   size=$(( $(cat /sys/class/block/$p/size) / 2048 )) # each block is 512-byte, convert to MB
			   echo -n "/dev/$p \"/dev/$p ($label) type $FS_TYPE size $size MB\" off " ;;
		esac
		
	done << EOF
$(ls /sys/class/block | grep -vE "^loop|^ram|^nbd|^sr")
EOF
)
	savedevice=$(eval "Xdialog --title \"$APPTITLE\" --wizard --no-tags --stdout --radiolist \"$msg\" 20 80 5 $items")
	next_wizard_step
}

### get the size and filename
choose_size_and_name() {
	local sizeinfo msg1 msg2 msg3 msg4 choice tmpmount minsize
	savedevicetype=$(guess_fstype $savedevice)

	# try mounting, also to get the free space
	tmpmount=$(mktemp -dp /tmp)
	if ! case $type in
		ntfs) ntfs-3g $savedevice $tmpmount	;;
		*) mount $savedevice $tmpmount ;;
	esac; then
		# failed to mount
		msg1="Problem: Unable to access $savedevice. Please choose another device/partition."
		Xdialog --title "$APPTITLE" --msgbox "$msg1" 0 0
		go_prev
		
		rmdir $tmpmount		
		return
	fi
	sizeinfo=$(df -h | grep $savedevice | awk '{print $2, " disk and has ", $4, "free space."; exit}')
	umount $tmpmount
	rmdir $tmpmount
	minsize=$(du -sm "$TMPFS_MOUNT" | awk '{print $1}') 
	
	msg1="$savedevice is a $sizeinfo"
	msg2="How large do you want your savefile? (in MB).\nYou need at least $minsize MB to save your current session."
	msg3="Specify the filename of the savefile (if in doubt, don't change)"
	msg4="Yes! I want to move Downloads folder to $savedevice too\nso it won't take up space in the savefile."
	
	if choice=$(Xdialog --title "$APPTITLE" --wizard --stdout --separator : --check "$msg4" $movedownload --2inputsbox "$msg1" 0 0 "$msg2" "$savesize" "$msg3" "$savefile"); then
		next_wizard_step
		set -- $choice
		savesize=${1%%:*}
		savefile=${1#*:}
		savepath=${savefile%/*}; [ "$savepath" = "$savefile" ] && savepath= # special case
		
		movedownload=on
		savedownload=$SAVE_DEV_PROTO/$savepath/${DOWNLOAD_FOLDER##*/} 		
		[ "$2" = "unchecked" ] && movedownload= && savedownload=
		
		# check that specified size must be more than current session size.
		if [ $savesize -lt $minsize ]; then
			Xdialog --title "$APPTITLE" --infobox "The specified size ($savesize MB) is less than the current session size ($minsize MB)." 0 0 10000
			go_prev 
		fi 
	else
		next_wizard_step	
	fi
}

### get filesystem type for savefile & whether encryption is used
choose_fs_and_encryption() {
	local choice msg1 msg2 items encstate
	msg1="Choose the filesystem you want to use for the savefile."
	msg2="Tick the box if you want to encrypt savefile."
	items=$(cat << EOF
ext4 "Ext4 - efficient for large savefile" off \
ext3 "Ext3 - resilient against crash" off \
ext2 "Ext2 - use this if using encryption" off
EOF
)
	[ "$encrypted" -a "$encrypted" != "none" ] && encstate=on
	if choice=$(eval "Xdialog --title \"$APPTITLE\" --wizard --stdout --no-tags --check \"$msg2\" \"$encstate\" --radiolist \"$msg1\" 0 0 0 $items"); then
		next_wizard_step
		set -- $choice; savefs=$1
		case $2 in
			checked)
				# if encrypted, add _crypt_ to savefile, and make sure savefs is always ext2
				encrypted=$ENCTYPE
				savefs=ext2				
				case $savefile in
					*_crypt_*) true ;; # already have _crypt_, do nothing
					*)	msg1=${savefile%%.*}; msg2=${savefile#*.}
						savefile=${msg1}_crypt_.${msg2}
						;;
				esac
				;;
			
			unchecked)
				# not encrypted, make sure filename doesn't contain _crypt_
				encrypted=none
				case $savefile in
					*_crypt_*)
						msg1=${savefile%%_crypt_*}; msg2=${savefile#*_crypt_}
						savefile=$msg1$msg2
					;;
				esac
				;;
		esac
	else
		next_wizard_step
	fi	
}

### get the password for encryption - if we use encryption
choose_password() {
	local msg msg1 msg2 choice
	case $encrypted in 
		none)
			next_wizard_step $prev_action
			;;
		*)
			while true; do
				msg="Please type in the password for your savefile. Password cannot contain spaces."
				msg1="Type your password."
				msg2="Type your password again - make sure it matches with the above."
				if choice=$(Xdialog --title "$APPTITLE" --backtitle "Savefile Password" --wizard --stdout --separator " " --password --password --2inputsbox "$msg" 0 0 "$msg1" "$savepassword" "$msg2" "$savepassword"); then
					# next action - must check password
					set -- $choice
					[ "$1" ] && [ "$1" = "$2" ] && savepassword="$1" && go_next && break

					msg="Problem: Passwords do not match. For your own safety, please re-type the passswords again."
					[ -z "$1" ] && msg="Problem: When using encryption, password cannot be blank."
					Xdialog --title "$APPTITLE" --backtitle "Savefile Password" --msgbox "$msg" 0 0
				else
					next_wizard_step
					break;
				fi
			done
			;;
	esac
}

### get a new hostname
choose_hostname() {
	local msg choice

	while true; do
		msg=$(cat <<EOF
The hostname is your computer's name on the network. \
The current hostname is chosen randomly during system start-up. \
You can change that, or just click OK to keep it.

Hostname should be unique across your entire network.
EOF
)
		if choice=$(Xdialog --fill --title "$APPTITLE" --backtitle "Set Hostname" --wizard --stdout --inputbox "$msg" 0 0 "$hostname"); then
			# next action - must check hostname is not blank
			[ "$choice" ] && hostname=$choice && go_next && break;

			msg="Problem: Hostname cannot be blank."
			Xdialog --title "$APPTITLE" --backtitle "Set Hostname" --msgbox "$msg" 0 0
		else
			next_wizard_step
			break
		fi
	done
}

### change root password
change_root_password() {
	local msg msg1 msg2 choice
	
	while true; do
		msg="Password for root user is currently set to \"$rootpassword\".\nYou can change that, or just click OK to keep it."
		msg1="Type your password. Password cannot contain spaces or colons."
		msg2="Type your password again - make sure it matches with the above."
		if choice=$(Xdialog --title "$APPTITLE" --backtitle "Set Root Password" --wizard --stdout --separator " " --password --password --2inputsbox "$msg" 0 0 "$msg1" "$rootpassword" "$msg2" "$rootpassword"); then
			# next action - must check password
			set -- $choice
			[ "$1" ] && [ "$1" = "$2" ] && rootpassword="$1" && go_next && break

			msg="Problem: Passwords do not match. For your own safety, please re-type the passswords again."
			[ -z "$1" ] && msg="Problem: root password cannot be blank."
			Xdialog --title "$APPTITLE" --backtitle "Set Root Password" --msgbox "$msg" 0 0
		else
			next_wizard_step
			break;
		fi
	done
}


### confirm all the selected parameter
confirmation() {
	local msg 

	msg=$(cat <<EOF
You have chosen to create a savefile as follows:

- Savefile location : $savedevice
- Path and filename : $savefile
- Size of savefile : $savesize MB
- Encryption : $encrypted $([ "$encrypted" -a "$encrypted" != "none" ] && echo with password: $savepassword) 
- New hostname: $hostname
- New root password: $rootpassword
$([ $movedownload ] && echo - Move Downloads folder to $savedevice)

Make sure you get everything correct.
If you click "Next", the savefile will be created and the hostname will be set.
If some of the above are wrong, click "Previous" now and correct them.
EOF
)
	Xdialog --left --title "$APPTITLE" --backtitle "Confirm Selected Settings" --wizard --yesno "$msg" 0 0
	next_wizard_step
}


### create the savefile - do the heavylifting here
create_savefile() {
	local msg xpid tmpmount devloop
	
	# tell user it will take a while
	msg="Creating savefile $savefile on $savedevice, please wait ..."
	Xdialog --title "$APPTITLE" --no-buttons --infobox "$msg" 0 0 3600000 & # one hour
	xpid=$!
	
	# check if it's already mounted. If it does, use existing mount point
	tmpmount=$(grep -m 1 $savedevice /proc/mounts | awk '{print $2}')
	if [ -z "$tmpmount" ]; then	
		tmpmount=$(mktemp -dp /tmp devsave.XXXXXX)
		if ! case $savedevicetype in
			ntfs) ntfs-3g $savedevice $tmpmount	;;
			*) mount $savedevice $tmpmount ;;
		esac; then
			# failed to mount
			kill $xpid
			msg="Problem: Unable to access $savedevice, please choose another device/partition."		
			Xdialog --title "$APPTITLE" --msgbox "$msg" 0 0
			go_step 2
			
			rmdir $tmpmount	
			return
		fi
	fi
		
	# check whether a savefile with that name already exist
	if [ -e $tmpmount/$savefile ]; then
		kill $xpid
		msg="Problem: $savefile on $savedevice already exist. Please choose a different name."
		Xdialog --title "$APPTITLE" --msgbox "$msg" 0 0 
		go_step 3
		
		umount $tmpmount && rmdir $tmpmount			
		return	
	fi
	
	# create the savefile
	[ "$savepath" ] && mkdir -p $tmpmount/$savepath
	if ! dd if=/dev/zero of=$tmpmount/$savefile count=0 bs=1 seek=${savesize}M; then
		kill $xpid	
		msg="Problem: Unable to create $savefile on $savedevice. Please choose another device/partition."	
		Xdialog --title "$APPTITLE" --msgbox "$msg" 0 0 
		go_step 2
		
		umount $tmpmount && rmdir $tmpmount
		return
	fi
	
	# make filesystem on the savefile - from here assume everything is successful
	savefile_mount=$(mktemp -dp /tmp pupsave.XXXXXX)
	case $encrypted in 
		none)
			mke2fs -F -m 0 -t $savefs $tmpmount/$savefile
			mount -o loop $tmpmount/$savefile $savefile_mount
			;;	
		*)
			modprobe cryptoloop
			modprobe $ENCTYPE
			devloop=$(losetup-FULL -f)
			echo "$savepassword" | losetup-FULL -p 0 -e $ENCTYPE $devloop $tmpmount/$savefile
			mke2fs -F -m 0 -t $savefs $devloop
			mount $devloop $savefile_mount
			;;
	esac
	echo SAVEFILE_MOUNT=$savefile_mount >> $BOOTSTATE_PATH
	
	# move download folder if requested
	if [ "$movedownload" ]; then
		cp -a $DOWNLOAD_FOLDER $tmpmount/$savepath
		rm -rf $DOWNLOAD_FOLDER
		ln -s $savedownload $DOWNLOAD_FOLDER
		chown -R spot:spot $tmpmount/$savepath/Downloads
		chown spot:spot $DOWNLOAD_FOLDER
	fi
	
	# change name and root password
	echo $hostname > /etc/hostname
	echo root:$rootpassword | chpasswd
	
	# done
	sync
	kill $xpid

	# keep savefile mounted, actual saving will be done by fatdog-merge-layers during shutdown ...
	msg="Done!\nSavefile $savefile has been created in $savedevice with encryption: $encrypted."
	[ "$movedownload" ] && msg="$msg\nDownloads folder moved to $savedownload."
	Xdialog --title "$APPTITLE" --infobox "$msg" 0 0 10000
	go_next
}

################## main ###################
while true; do 
	case $step in 
		1) introduction ;;
		2) choose_device ;;
		3) choose_size_and_name ;;
		4) choose_fs_and_encryption ;;
		5) choose_password ;;
		6) choose_hostname ;;
		7) change_root_password ;;
		8) confirmation ;;
		9) create_savefile ;;
		10|exit) break ;;
		warning) data_loss_warning ;;
	esac
done
